package com.perforce.hws.p4base;

import com.perforce.hwsclient.models.ChangelistAction;
import com.perforce.hwsclient.models.ChangelistRequest;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.server.IServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Base64;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static java.nio.file.Files.createDirectories;
import static java.nio.file.Files.write;
import static java.nio.file.StandardOpenOption.CREATE;

/**
 * This is logic that helps to create and submit a changelist.
 */
public class CreateChange implements
        Function<ChangelistRequest, List<ResultMap>>,
        ChangelistMethods,
        P4Methods,
        PathUtils,
        TemporaryClient,
        WhereMethods {

    private static final Logger logger = LoggerFactory.getLogger(CreateChange.class);

    private Supplier<IServer> serverSupplier;

    // Internal container of the output of "p4 files" on each of our files to
    // operate on.
    private Map<String, ResultMap> depotPathToFile = new HashMap<>();

    // Output results for function
    private List<ResultMap> results = null;

    public CreateChange(Supplier<IServer> serverSupplier) {
        this.serverSupplier = serverSupplier;
    }

    @Override
    public List<ResultMap> apply(ChangelistRequest request) {
        try {
            if (request.getStream() != null) {
                withStream(serverSupplier, request.getStream(), r -> submitFiles(request));
            } else {
                withClient(serverSupplier, r -> submitFiles(request));
            }

            // Note: This exception handling is actually not very useful here,
            // since our submitFiles method will do it's own exception handling.
            // So take care in case you add actual exception handling.

        } catch (P4JavaException e) {
            throw new P4JavaRuntimeException(e);
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
        return results;
    }

    private void submitFiles(ChangelistRequest request) {
        String change = null;

        try {
            change = createChangelist(serverSupplier, request.getDescription());
            fetchDepotPathToFile(request);

            final String ch = change;
            request.getActions().stream().forEach(a -> executeAction(request, ch, a));

            results = exec(serverSupplier, "submit", "-c", change);

        } catch (Throwable th) {
            // Delete the changelist
            if (change != null) {
                try {
                    exec(serverSupplier, "revert", "-c", change, "//...");
                    exec(serverSupplier, "change", "-d", change);
                } catch (Exception captured) {
                    logger.info("problem deleting changelist " + change, captured);
                }
            }

            // This will basically bypass any exception handling in the apply()
            // method.
            throw new RuntimeException(th);
        }
    }

    private void fetchDepotPathToFile(ChangelistRequest request) {

        List<String> depotPaths = depotPathsForRequest(request);

        List<ResultMap> filesResults =
                exec(serverSupplier, "files", depotPaths.toArray(new String[0]));

        filesResults.stream().forEach(map -> {
            String depotFile = (String) map.get("depotFile");
            if (depotPaths.contains(depotFile)) {
                depotPathToFile.put(depotFile, map);
            }
        });
    }

    private List<String> depotPathsForRequest(ChangelistRequest request) {
        if (request.getActions() == null) {
            return null;
        }
        return request.getActions().stream()
                .map(a -> depotPathForAction(request, a))
                .collect(Collectors.toList());
    }

    private String depotPathForAction(ChangelistRequest request, ChangelistAction action) {
        if (request.getStream() != null) {
            return String.format("%s/%s", request.getStream(), action.getDepotFile());
        } else {
            return action.getDepotFile();
        }
    }

    private String fromDepotPathForAction(ChangelistRequest request, ChangelistAction action) {
        if (request.getStream() != null) {
            return String.format("%s/%s", request.getStream(), action.getFromDepotFile());
        } else {
            return action.getFromDepotFile();
        }
    }

    private void executeAction(ChangelistRequest request,
                               String change,
                               ChangelistAction action) {
        try {
            switch (action.getActionType()) {
                case "upload":
                    uploadFile(request, change, action);
                    break;

                case "branch":
                    integrateFile(request, change, action);
                    break;

                case "move":
                    moveFile(request, change, action);
                    break;

                case "delete":
                    deleteFile(request, change, action);
            }
        } catch (IOException io) {
            throw new IllegalStateException(io);
        }
    }

    private void uploadFile(ChangelistRequest request,
                            String change,
                            ChangelistAction action) throws IOException {
        String depotPath = depotPathForAction(request, action);
        if (hasSpecialChars(depotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + depotPath);
        }

        if (action.getRequireVersion() != null && action.getRequireVersion() >= 0) {
            if (action.getRequireVersion() == 0 && pathExists(depotPath)) {
                throw new P4JavaRuntimeException("depotPath exists: " + depotPath);
            }

            ResultMap fileInfo = depotPathToFile.get("depotPath");

            int current = 0;
            if (pathExists(depotPath)) {
                current = Integer.parseInt((String) fileInfo.get("rev"));
            }
            if (current != action.getRequireVersion()) {
                throw new P4JavaRuntimeException("current version '" +
                        current + "' is not required version '" +
                        action.getRequireVersion() + "'");
            }
        }

        String localPath = execWhere(serverSupplier, depotPath);

        if (pathExists(depotPath)) {
            exec(serverSupplier, "sync", depotPath);
            exec(serverSupplier, "edit", "-c", change, localPath);
            saveContent(depotPath, action);
        } else {
            saveContent(depotPath, action);
            exec(serverSupplier, "add", "-c", change, localPath);
        }
    }

    private void integrateFile(ChangelistRequest request,
                               String change,
                               ChangelistAction action) {
        String depotPath = depotPathForAction(request, action);
        if (hasSpecialChars(depotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + depotPath);
        }

        String fromDepotPath = fromDepotPathForAction(request, action);
        if (hasSpecialChars(fromDepotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + fromDepotPath);
        }

        exec(serverSupplier, "integ", "-c", change, fromDepotPath, depotPath);
    }

    private void moveFile(ChangelistRequest request,
                          String change,
                          ChangelistAction action) {
        String depotPath = depotPathForAction(request, action);
        if (hasSpecialChars(depotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + depotPath);
        }

        String fromDepotPath = depotPathForAction(request, action);
        if (hasSpecialChars(fromDepotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + fromDepotPath);
        }

        exec(serverSupplier, "move", "-c", change, fromDepotPath, depotPath);
    }

    private void deleteFile(ChangelistRequest request,
                            String change,
                            ChangelistAction action) {
        String depotPath = depotPathForAction(request, action);
        if (hasSpecialChars(depotPath)) {
            throw new IllegalArgumentException("Special characters are not " +
                    "allowed: " + depotPath);
        }

        exec(serverSupplier, "sync", depotPath);
        exec(serverSupplier, "delete", "-c", change, depotPath);
    }

    private void saveContent(String depotPath, ChangelistAction action) throws IOException {
        String localPath = execWhere(serverSupplier, depotPath);
        byte bytes[] = Base64.getDecoder().decode(action.getContent());
        Path path = Paths.get(localPath);
        createDirectories(path.getParent());
        write(Paths.get(localPath), bytes, CREATE);
    }

    private boolean pathExists(String depotPath) {
        ResultMap fileInfo = depotPathToFile.get(depotPath);
        if (fileInfo != null &&
                !"delete".equals(fileInfo.get("action"))) {
            return true;
        }
        return false;
    }


}
