package com.perforce.hws.p4base;

import com.perforce.p4java.exception.*;
import com.perforce.p4java.impl.mapbased.rpc.RpcServer;
import com.perforce.p4java.impl.mapbased.server.Server;
import com.perforce.p4java.server.IServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Supplier;
import java.util.stream.Collectors;

/**
 * Syntactic sugar around the P4Java API methods that we use extensively.
 * <p>
 */
public interface P4Methods {

    Logger logger = LoggerFactory.getLogger(P4Methods.class);

    /**
     * Execute a command with no input.
     *
     * @param command
     * @param args
     */
    default List<ResultMap> exec(Supplier<IServer> supplier,
                                 String command,
                                 String... args) {
        IServer s = supplier.get();
        if (logger.isDebugEnabled()) {
            logger.debug("command: " + command +
                    ", args: " + Arrays.toString(args));
        }
        try {
            return toResultMap(s.execMapCmd(command, args, null));
        } catch (ConnectionException | RequestException | AccessException e) {
            throw new P4JavaRuntimeException(e);
        }
    }

    /**
     * Execute a command, specifying input as a Map.
     *
     * @param input
     * @param command
     * @param args
     * @return
     */
    default List<ResultMap> execInput(Supplier<IServer> supplier,
                                      Map<String, Object> input,
                                      String command,
                                      String... args) {
        IServer s = supplier.get();
        if (logger.isDebugEnabled()) {
            if (command.equals("passwd") || command.equals("login")) {
                logger.debug("command: " + command +
                        ", args: " + Arrays.toString(args));
            } else {
                logger.debug("command: " + command +
                        ", args: " + Arrays.toString(args) +
                        ", input: " + input);
            }
        }
        if (command.equals("passwd")) {
            // The RpcServer implementation (always in use) in P4Java can
            // improperly cache keys for too long.
            ((RpcServer) s).setSecretKey(s.getUserName(), null);
        }
        try {
            return toResultMap(s.execMapCmd(command, args, input));
        } catch (ConnectionException | RequestException | AccessException e) {
            throw new P4JavaRuntimeException(e);
        }
    }

    /**
     * A workaround method to the occasional limitation on p4java that blocks
     * some kinds commands that use map for input.
     * <p>
     * Don't use this unless you really have to, it's a road far less travelled.
     *
     * @return
     */
    default List<ResultMap> execInputString(Supplier<IServer> supplier,
                                            String input,
                                            String command,
                                            String... args) {
        IServer s = supplier.get();
        if (logger.isDebugEnabled()) {
            logger.debug("command: " + command +
                    ", args: " + Arrays.toString(args) +
                    ", input: " + input);
        }
        // This error handling is very unclear compared to execMapCmd.
        try {
            Server internal = (Server)s;
            List<Map<String, Object>> raw =
                    internal.execInputStringMapCmdList(command, args, input);
            return toResultMap(raw.toArray(new HashMap[0]));
        } catch (P4JavaException e) {
            throw new P4JavaRuntimeException(e);
        }
    }

    /**
     * Converts our typical output maps to a ResultMap list.
     * <br/>
     * Note: server errors are often returned as Maps too. Each ResultMap will
     * need to be checked if it's an error or not (there's usually just one).
     *
     * @param results Internal hash maps we convert to ResultMap
     * @return Converted list of ResultMap instances.
     * @throws ServerErrorException, ServerFailureException if the input results indicate some kind of error.
     */
    default List<ResultMap> toResultMap(Map<String, Object>[] results) {
        List<ResultMap> maps = Arrays.stream(results)
                .map((m) -> new ResultMap(m))
                .collect(Collectors.toList());

        // Check for server failure, errors, or warnings in the results. Our
        // "low level" p4java commands do not check for this at all.
        maps.forEach((m) -> {
            int maxSeverity = m.getMaxSeverity();
            if (maxSeverity >= getFatalSeverityLevel()) {
                logger.warn("Fatal error from p4d: {}", m);
                throw new ServerFailureException(m);
            } else if (maxSeverity >= getErrorSeverityLevel()) {
                logger.warn("Error from p4d: {}", m);
                throw new ServerErrorException(m);
            } else if (maxSeverity >= getWarningSeverityLevel()) {
                if (logger.isWarnEnabled()) {
                    logger.warn(m.toString());
                }
            } else {
                if (logger.isDebugEnabled()) {
                    logger.debug(m.toString());
                }
            }
        });

        return maps;
    }

    /**
     * If we spot a fatal severity, the result map will be thrown as a
     * ServerFailureException.
     *
     * @return Severity level to treat as fatal. Defaults to E_FATAL.
     */
    default int getFatalSeverityLevel() {
        return MessageSeverityCode.E_FATAL;
    }

    /**
     * If we spot an error severity, we throw the ResultMap as a
     * ServerErrorException.
     *
     * @return Severity level to interpet as error. Defaults to E_FAILED.
     */
    default int getErrorSeverityLevel() {
        return MessageSeverityCode.E_FAILED;
    }

    /**
     * If we get a warning severity, we log these messages as a warning.
     * <p>
     * Everything else will either be an error or a debug message.
     *
     * @return Severity level to treat as warnings, defaults to E_WARN
     */
    default int getWarningSeverityLevel() {
        return MessageSeverityCode.E_WARN;
    }

    /**
     * Runs the "p4 passwd" command on the current connection.
     *
     * @param supplier    IServer handle
     * @param oldPassword Current password, can be null if there isn't one
     * @param newPassword The password to change to
     * @return Our typical result map
     * @throws P4JavaException
     */
    default List<ResultMap> changePassword(Supplier<IServer> supplier,
                                           String oldPassword,
                                           String newPassword) throws P4JavaException {
        Map<String, Object> input = new HashMap<>();
        input.put("oldPassword", oldPassword);
        input.put("newPassword", newPassword);
        input.put("newPassword2", newPassword);
        return execInput(supplier, input, "passwd");
    }
}
