package com.perforce.hws.p4base;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import static java.lang.Integer.compare;
import static java.lang.Integer.valueOf;
import static java.lang.String.format;
import static java.util.stream.Collectors.toList;

/**
 * Utilities to manipulate ResultMap instances.
 */
public interface ResultMapMethods {

    Pattern INDEXED_KEY = Pattern.compile("(?<key>[^\\d]+)(?<index>\\d+)$");

    /**
     * Convert "array-like" key->value entries to a single array.
     * <p>
     * P4Java will return "array-like" values in a "indexed key" format.
     * This will look for any entry that ends with a number and will return
     * a new map that contains a single key -> array value.
     * <p>
     * For example, if the result map contains "View0" -> a, "View1" -> b, this
     * will return a map with the entry "View" -> [a, b] in it's place.
     *
     * @param resultMap
     * @return
     */
    default Map<String, Object> collateArrayEntries(ResultMap resultMap) {
        HashMap<String, Object> converted = new HashMap<>();

        // Get list of "index-able" keys. We sort each key by the index value.
        // We end up with an array of [Original, KeyName, IndexValue]
        List<String[]> indexable = resultMap.entrySet().stream()
                .map(e -> INDEXED_KEY.matcher(e.getKey()))
                .filter(m -> m.matches())
                .sorted((m, n) -> compare(valueOf(m.group("index")), valueOf(n.group("index"))))
                .map(m -> new String[]{m.group(), m.group("key")})
                .collect(toList());

        // Copy each indexable key into our converted map
        indexable.forEach(arr -> {
            if (!converted.containsKey(arr[1])) {
                converted.put(arr[1], new ArrayList<>());
            }
            List list = (List) converted.get(arr[1]);
            list.add(resultMap.get(arr[0]));
        });

        // Copy everything else that isn't "index-able".
        resultMap.entrySet().stream()
                .filter(e -> !INDEXED_KEY.matcher(e.getKey()).matches())
                .forEach(e -> converted.put(e.getKey(), e.getValue()));

        return converted;
    }

    /**
     * Converts array values into "keyed" index values that p4java will send
     * back over the wire.
     * <p>
     * If the "collated" map contains a key "Foo" -> ["a", "b"] this will return
     * a map with keys "Foo0" -> "a", "Foo1" -> "b".
     * <p>
     * The collated value is expected to be a List type.
     *
     * @param collated
     * @return
     */
    default Map<String, Object> divideArrayEntries(Map<String, Object> collated) {
        HashMap<String, Object> divided = new HashMap<>();

        collated.forEach((k, v) -> {
            if (v instanceof List) {
                List l = (List) v;
                for (int index = 0; index < l.size(); ++index) {
                    Object lv = l.get(index);
                    divided.put(format("%s%d", k, index), lv);
                }
            } else {
                divided.put(k, v);
            }
        });

        return divided;
    }
}
