package com.perforce.hws.p4base;

import com.perforce.hws.util.FilesUtils;
import com.perforce.hws.util.StringUtils;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.server.IServer;

import java.io.IOException;
import java.nio.file.Path;
import java.util.List;
import java.util.function.Supplier;

/**
 * Used to indicate your class needs a TemporaryClient
 */
public interface TemporaryClient extends P4Methods, FilesUtils, StringUtils {

    /**
     * Your applications will implement these, most likely as closures, to
     * execute using the server bound to the particular local temporary
     * directory. <p></p>
     * <p>
     * The client will be deleted and the rootDir will be cleaned up upon
     * exit or any failure of the application.
     */
    @FunctionalInterface
    interface UseClient {

        /**
         * @param root
         * @throws P4JavaException
         * @throws IOException
         * @throws P4BaseException
         */
        void use(String root) throws P4JavaException, IOException, P4BaseException;
    }

    /**
     * Execute methods using a temporary wide-open "classic" client.
     *
     * @param serverSupplier
     * @param useClient
     * @throws P4JavaException
     * @throws IOException
     */
    default void withClient(Supplier<IServer> serverSupplier, UseClient useClient)
            throws P4JavaException, IOException, P4BaseException {

        IServer server = serverSupplier.get();

        if (!server.isConnected()) {
            server.connect();
        }
        String name = randomString();
        Path rootPath = createTempDirectory();
        try {
            // Create our default client spec here.
            List<ResultMap> results = exec(serverSupplier, "client", "-o", name);
            ResultMap client = results.get(0);
            client.put("Root", rootPath.toAbsolutePath().toString());

            client.put("Options", "allwrite clobber nocompress unlocked nomodtime rmdir");

            // If there's only one depot in the server, the generated view
            // will only be "//depot/... //view/..." instead of
            // "//depot/... //view/depot/..." like it is for servers with
            // multiple depots. We need to revise the view here just to
            // remain consistent.
            if (client.indexedEntries("View").size() == 1) {
                client.put("View0", "//depot/... //" + name + "/depot/...");
            }

            // The "-x" flag is a "super undocumented" feature that will fail
            // if the client already exists. It's often used as a means for
            // creating locks across servers. Our client names are pretty safely
            // random, but, you never know.
            //
            // This is known to cause problems with 14.2 servers that are not
            // patched.
            execInput(serverSupplier, client, "client", "-i", "-x");

            // Try to satisfy P4Java's insistence of using lots of heavyweight
            // cached objects everywhere. Part of me things we could just use
            // "setClientName" but ... better safe than sorry
            IClient c = server.getClient(name);
            server.setCurrentClient(c);

            useClient.use(rootPath.toString());
        } finally {
            P4JavaRuntimeException handledEx = null;
            try {
                exec(serverSupplier, "client", "-d", name);
            } catch (P4JavaRuntimeException ex) {
                handledEx = ex;
            }

            rmtree(rootPath.toString());

            if (handledEx != null) {
                throw handledEx;
            }
        }
    }

    /**
     * Execute methods using a temporary stream-based client.
     *
     * @param serverSupplier Our handle to the P4Java server
     * @param stream The stream ID, e.g. <code>//my_stream/main</code>
     * @param useClient The function we call using the temporary client
     * @throws P4JavaException
     * @throws IOException
     * @throws P4BaseException
     */
    default void withStream(Supplier<IServer> serverSupplier,
                            String stream,
                            UseClient useClient)
        throws P4JavaException, IOException, P4BaseException {

        IServer server = serverSupplier.get();

        // This is frequently necessary in test environments
        if (!server.isConnected()) {
            server.connect();
        }

        String name = randomString();
        Path rootPath = createTempDirectory();

        try {
            List<ResultMap> streamResults = exec(serverSupplier,
                    "client", "-o", "-S", stream, name);

            ResultMap client = streamResults.get(0);
            client.put("Root", rootPath.toAbsolutePath().toString());

            client.put("Options", "allwrite clobber nocompress unlocked nomodtime rmdir");

            // The "-x" flag is a "super undocumented" feature that will fail
            // if the client already exists. It's often used as a means for
            // creating locks across servers. Our client names are pretty safely
            // random, but, you never know.
            //
            // This is known to cause problems with 14.2 servers that are not
            // patched.
            execInput(serverSupplier, client, "client", "-i", "-x");

            // Try to satisfy P4Java's insistence of using lots of heavyweight
            // cached objects everywhere. Part of me things we could just use
            // "setClientName" but ... better safe than sorry
            IClient c = server.getClient(name);
            server.setCurrentClient(c);

            useClient.use(rootPath.toString());

        } finally {
            P4JavaRuntimeException handledEx = null;
            try {
                exec(serverSupplier, "client", "-d", name);
            } catch (P4JavaRuntimeException ex) {
                handledEx = ex;
            }

            rmtree(rootPath.toString());

            if (handledEx != null) {
                throw handledEx;
            }
        }
    }
}
